#include <avr/io.h>
#include <avr/interrupt.h>

#include "cpu.h"
#include "xitoa.h"
#include "serx.h"
#include "timer2.h"
#include "schedule.h"

/* ganz ohne statemachine, ablauf in mehrern phasen demo. */
static volatile u8 comm_state=0;
void comm(void)
{
    writeln("comm");
    while(1) {
        while( comm_state == 0 ) schedule();
        comm_state=0;
        writeln("state 1");

        while( comm_state == 0 ) schedule();
        writeln("state 2");
        comm_state=0;

        while( comm_state == 0 ) schedule();
        writeln("state 3");
        comm_state=0;
    }
}

/* anzahl task-wechsel pro SEKUNDE ausgeben */
void rpm(void)
{
    u32 rpm = 0;
    u32 tm;
    while(1) {
        tm = timer2_get();
        while( timer2_elapsed_time(tm) < 2000 ) {
            rpm++;
            schedule();
        }
        writeln( "Rounds per Second: %lu", rpm  / 2);
        rpm=0;
    }

}


/* jede sekunde den task "comm" weiterlaufen lassen */
void task_main(void)
{
    u32 tm;
    writeln("main");
    while(1) {
        tm = timer2_get();
        writeln("task main");
        if( comm_state == 0 ) comm_state++;
        schedule();
        while( timer2_elapsed_time(tm) < 1000 ) schedule();
        XBI_PORT(PIN_LED);
    }
}


/* create debug interface, prefix is "log":
   PORT  : USART-0
   IO-BUF: 2^7 = 128
   functions: log_in_is_empty, log_in_get;
*/
GEN_SERX(log, 0, 128)
static void ser_init(void)
{
  log_init(REG_UBRR0);
  UCSR0A = REG_UCSR0A;
  xfunc_out = log_put;   // stdout for xputs,xprintf
}

int main(void)
{
    cfg_lo(PIN_LED);
    ser_init();
    timer2_init();
    sei();
    task_create(comm);
    task_create(task_main);
    task_create(rpm);
    run(); /* never returns */
}
